/* Copyright (c) 1997-2003 Miller Puckette.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution.

iemlib2 written by Thomas Musil (c) IEM KUG Graz Austria 2000 - 2003 */

#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif

#include "m_imp.h"
#include "g_canvas.h"
#include "iemlib.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

EXTERN void canvas_getargs(int *argcp, t_atom **argvp);

/* ------------------------ iem_anything_kernel ---------------------------- */
static t_class *iem_anything_kernel_class;

typedef struct _iem_anything_kernel
{
	t_object			x_obj;
	int						x_inlet_select;
	int						x_size;
	int						x_ac;
	t_atom				*x_at;
	t_symbol			*x_sym;
	t_atomtype		x_type;
} t_iem_anything_kernel;

static void iem_anything_kernel_atcopy(t_atom *src, t_atom *dst, int n)
{
	while(n--)
		*dst++ = *src++;
}

static void iem_anything_kernel_inlet_select(t_iem_anything_kernel *x, t_floatarg in_sel_01)
{
	x->x_inlet_select = (int)in_sel_01;
}

static void iem_anything_kernel_bang(t_iem_anything_kernel *x)
{
	if(x->x_inlet_select)
	{
		x->x_type = A_NULL;
	}
	else
	{
		if(x->x_type == A_FLOAT)
			outlet_float(x->x_obj.ob_outlet, atom_getfloat(x->x_at));
		else if(x->x_type == A_GIMME)
			outlet_list(x->x_obj.ob_outlet, &s_list, x->x_ac, x->x_at);
		else if(x->x_type == A_COMMA)
			outlet_anything(x->x_obj.ob_outlet, x->x_sym, x->x_ac, x->x_at);
		else if(x->x_type == A_SYMBOL)
			outlet_symbol(x->x_obj.ob_outlet, atom_getsymbol(x->x_at));
		else if(x->x_type == A_NULL)
			outlet_bang(x->x_obj.ob_outlet);
		else if(x->x_type == A_POINTER)
			outlet_pointer(x->x_obj.ob_outlet, (t_gpointer *)x->x_at->a_w.w_gpointer);
	}
}

static void iem_anything_kernel_float(t_iem_anything_kernel *x, t_float f)
{
	x->x_ac = 1;
	SETFLOAT(x->x_at, f);
	x->x_type = A_FLOAT;
	if(!x->x_inlet_select)
		outlet_float(x->x_obj.ob_outlet, f);
}

static void iem_anything_kernel_symbol(t_iem_anything_kernel *x, t_symbol *s)
{
	x->x_ac = 1;
	SETSYMBOL(x->x_at, s);
	x->x_type = A_SYMBOL;
	if(!x->x_inlet_select)
		outlet_symbol(x->x_obj.ob_outlet, s);
}

static void iem_anything_kernel_pointer(t_iem_anything_kernel *x, t_gpointer *gp)
{
	x->x_ac = 1;
	SETPOINTER(x->x_at, gp);
	x->x_type = A_POINTER;
	if(!x->x_inlet_select)
		outlet_pointer(x->x_obj.ob_outlet, gp);
}

static void iem_anything_kernel_list(t_iem_anything_kernel *x, t_symbol *s, int ac, t_atom *av)
{
	if(ac > x->x_size)
	{
		x->x_at = (t_atom *)resizebytes(x->x_at, x->x_size*sizeof(t_atom), ac*sizeof(t_atom));
		x->x_size = ac;
	}
	x->x_ac = ac;
	x->x_sym = &s_list;
	iem_anything_kernel_atcopy(av, x->x_at, ac);
	x->x_type = A_GIMME;
	if(!x->x_inlet_select)
		outlet_list(x->x_obj.ob_outlet, s, ac, av);
}

static void iem_anything_kernel_anything(t_iem_anything_kernel *x, t_symbol *s, int ac, t_atom *av)
{
	if(ac > x->x_size)
	{
		x->x_at = (t_atom *)resizebytes(x->x_at, x->x_size*sizeof(t_atom), ac*sizeof(t_atom));
		x->x_size = ac;
	}
	x->x_ac = ac;
	x->x_sym = s;
	iem_anything_kernel_atcopy(av, x->x_at, ac);
	x->x_type = A_COMMA;
	if(!x->x_inlet_select)
		outlet_anything(x->x_obj.ob_outlet, s, ac, av);
}

static void iem_anything_kernel_free(t_iem_anything_kernel *x)
{
	if(x->x_at)
		freebytes(x->x_at, x->x_size * sizeof(t_atom));
}

static void *iem_anything_kernel_new(void)
{
	t_iem_anything_kernel *x = (t_iem_anything_kernel *)pd_new(iem_anything_kernel_class);
	t_glist *glist = (t_glist *)canvas_getcurrent();
	t_canvas *canvas=glist_getcanvas(glist);
	int ac=0;
	t_atom *av;

	canvas_setcurrent(canvas);
	canvas_getargs(&ac, &av);
	canvas_unsetcurrent(canvas);

	x->x_type = A_NULL;
	if(!ac)
	{
		x->x_type = A_NULL;
		x->x_sym = &s_bang;
		x->x_size = 10;
		x->x_ac = 0;
		x->x_at = (t_atom *)getbytes(x->x_size * sizeof(t_atom));
	}
	else if(ac == 1)
	{
		if(IS_A_SYMBOL(av,0))
		{
			x->x_type = A_COMMA;
			x->x_sym = atom_getsymbol(av);
			x->x_size = 10;
			x->x_ac = 0;
			x->x_at = (t_atom *)getbytes(x->x_size * sizeof(t_atom));
		}
		else
		{
			if(IS_A_FLOAT(av,0))
			{
				x->x_type = A_FLOAT;
				x->x_sym = &s_float;
			}
			else if(IS_A_POINTER(av,0))
			{
				x->x_type = A_POINTER;
				x->x_sym = &s_pointer;
			}
			x->x_size = 10;
			x->x_ac = 1;
			x->x_at = (t_atom *)getbytes(x->x_size * sizeof(t_atom));
			x->x_at[0] = *av;
		}
	}
	else /* ac > 1 */
	{
		if(IS_A_SYMBOL(av,0))
		{
			x->x_type = A_COMMA;/*for anything*/
			x->x_sym = atom_getsymbol(av++);
			ac--;
		}
		else
		{
			x->x_type = A_GIMME;
			x->x_sym = &s_list;
		}
		if(ac < 10)
			x->x_size = 10;
		else
			x->x_size = ac;
		x->x_ac = ac;
		x->x_at = (t_atom *)getbytes(x->x_size * sizeof(t_atom));
		iem_anything_kernel_atcopy(av, x->x_at, ac);
	}
	x->x_inlet_select = 1;
	outlet_new(&x->x_obj, &s_list);
	return (x);
}

void iem_anything_kernel_setup(void)
{
	char str[2];

	str[0] = 1;/*inlet-sym = "\0x01"*/
	str[1] = 0;
	iem_anything_kernel_class = class_new(gensym("iem_anything_kernel"),
		(t_newmethod)iem_anything_kernel_new, (t_method)iem_anything_kernel_free,
																						sizeof(t_iem_anything_kernel), 0, 0);
	class_addbang(iem_anything_kernel_class, (t_method)iem_anything_kernel_bang);
	class_addfloat(iem_anything_kernel_class, (t_method)iem_anything_kernel_float);
	class_addsymbol(iem_anything_kernel_class, iem_anything_kernel_symbol);
	class_addpointer(iem_anything_kernel_class, iem_anything_kernel_pointer);
	class_addlist(iem_anything_kernel_class, iem_anything_kernel_list);
	class_addmethod(iem_anything_kernel_class, (t_method)iem_anything_kernel_inlet_select, gensym(str), A_FLOAT, 0);
	class_addanything(iem_anything_kernel_class, iem_anything_kernel_anything);
	class_sethelpsymbol(iem_anything_kernel_class, gensym("iemhelp/help-iem_anything"));
}

