/* Copyright (c) 1997-2003 Miller Puckette.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution.

iemlib2 written by Thomas Musil (c) IEM KUG Graz Austria 2000 - 2003 */

#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif

#include "m_pd.h"
#include "iemlib.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

/* ------------------------ prepend ---------------------------- */
static t_class *prepend_class;

typedef struct _prepend
{
    t_object   x_obj;
    int        x_n;
    int        x_ac;
    int        x_head_ac;
    t_atom     *x_at;
    t_symbol   *x_sym;
    t_atomtype x_type;
} t_prepend;

static void prepend_out(t_prepend *x)
{
    if(x->x_type == A_COMMA)
        outlet_anything(x->x_obj.ob_outlet, x->x_sym, x->x_ac, x->x_at);
    else if(x->x_type == A_GIMME)
        outlet_list(x->x_obj.ob_outlet, &s_list, x->x_ac, x->x_at);
    else if(x->x_type == A_FLOAT)
        outlet_float(x->x_obj.ob_outlet, atom_getfloat(x->x_at));
    else if(x->x_type == A_SYMBOL)
        outlet_symbol(x->x_obj.ob_outlet, atom_getsymbol(x->x_at));
    else if(x->x_type == A_NULL)
        outlet_bang(x->x_obj.ob_outlet);
    else if(x->x_type == A_POINTER)
        outlet_pointer(x->x_obj.ob_outlet, (t_gpointer *)x->x_at->a_w.w_gpointer);
}

/*static void prepend_bang(t_prepend *x)
{                                       
    outlet_anything(x->x_obj.ob_outlet, x->x_sym, x->x_ac, x->x_at);
}*/

static void prepend_float(t_prepend *x, t_float f)
{
    t_atom *at;

    SETFLOAT(&x->x_at[x->x_head_ac], f);
    x->x_ac = x->x_head_ac + 1;
    prepend_out(x);
    /*outlet_anything(x->x_obj.ob_outlet, x->x_sym, x->x_ac, x->x_at);*/
}

static void prepend_symbol(t_prepend *x, t_symbol *s)
{
    t_atom *at;

    SETSYMBOL(&x->x_at[x->x_head_ac], s);
    x->x_ac = x->x_head_ac + 1;
    prepend_out(x);
    /*outlet_anything(x->x_obj.ob_outlet, x->x_sym, x->x_ac, x->x_at);*/
}

/*static void prepend_pointer(t_prepend *x, t_gpointer *gp)
{
    if(!x->x_at)
    {
        x->x_n = 1;
	x->x_at = (t_atom *)getbytes(sizeof(t_atom));
    }
    x->x_ac = 1;
    SETPOINTER(x->x_at, gp);
    x->x_sym = &s_pointer;
    outlet_pointer(x->x_obj.ob_outlet, gp);
}*/

static void prepend_list(t_prepend *x, t_symbol *s, int ac, t_atom *av)
{
    int i, j, hac;

    if((ac + x->x_head_ac) > x->x_n)
    {
	t_atom *at;

        at = (t_atom *)getbytes((ac + x->x_head_ac + 2)*sizeof(t_atom));
        hac = x->x_head_ac;
        for(i=0; i<hac; i++)
	    at[i] = x->x_at[i];
        if(x->x_at)
            freebytes(x->x_at, x->x_n * sizeof(t_atom));
        x->x_n = ac + hac + 2;
	x->x_at = at;
    }
    hac = x->x_ac = x->x_head_ac + ac;
    for(i=x->x_head_ac, j=0; i<hac; i++, j++)
	x->x_at[i] = av[j];
    prepend_out(x);
    /*outlet_anything(x->x_obj.ob_outlet, x->x_sym, x->x_ac, x->x_at);*/
}

static void prepend_anything(t_prepend *x, t_symbol *s, int ac, t_atom *av)
{
    int i, j, hac;

    if((ac + x->x_head_ac + 1) > x->x_n)
    {
	t_atom *at;

	at = (t_atom *)getbytes((ac + x->x_head_ac + 3)*sizeof(t_atom));
        hac = x->x_head_ac;
        for(i=0; i<hac; i++)
	    at[i] = x->x_at[i];
        if(x->x_at)
            freebytes(x->x_at, x->x_n * sizeof(t_atom));
        x->x_n = ac + hac + 3;
	x->x_at = at;
    }
    SETSYMBOL(&x->x_at[x->x_head_ac], s);
    hac = x->x_ac = x->x_head_ac + ac + 1;
    for(i=x->x_head_ac+1, j=0; i<hac; i++, j++)
	x->x_at[i] = av[j];
    prepend_out(x);
    /*outlet_anything(x->x_obj.ob_outlet, x->x_sym, x->x_ac, x->x_at);*/
}

static void prepend_free(t_prepend *x)
{
    if(x->x_at)
        freebytes(x->x_at, x->x_n * sizeof(t_atom));
}

static void *prepend_new(t_symbol *s, int ac, t_atom *av)
{
    t_prepend *x = (t_prepend *)pd_new(prepend_class);
    int i;

    if(!ac)
    {
	post("prepend-ERROR: must have at least one argument!");
        x->x_at = (t_atom *)0;
        return(0);
    }
    else
    {
	if(IS_A_SYMBOL(av,0))
	{
	    x->x_sym = atom_getsymbol(av++);
	    x->x_type = A_COMMA;
            ac--;
	}
	else if(IS_A_FLOAT(av,0))
	{
	    x->x_sym = &s_list;
            x->x_type = A_GIMME;
	}
	x->x_head_ac = ac;
        x->x_ac = x->x_head_ac;
        x->x_n = x->x_head_ac + 2;
        x->x_at = (t_atom *)getbytes(x->x_n*sizeof(t_atom));
        for(i=0; i<ac; i++)
	    x->x_at[i] = *av++;
    }
    outlet_new(&x->x_obj, &s_list);
    return (x);
}

void prepend_setup(void)
{
    prepend_class = class_new(gensym("prepend"), (t_newmethod)prepend_new,
			       (t_method)prepend_free, sizeof(t_prepend), 0, A_GIMME, 0);
    class_addcreator((t_newmethod)prepend_new, gensym("pp"), A_GIMME, 0);
    /*class_addbang(prepend_class, (t_method)prepend_bang);*/
    class_addanything(prepend_class, prepend_anything);
    class_addlist(prepend_class, prepend_list);
    /*class_addpointer(prepend_class, prepend_pointer);*/
    class_addfloat(prepend_class, (t_method)prepend_float);
    class_addsymbol(prepend_class, prepend_symbol);
    class_sethelpsymbol(prepend_class, gensym("iemhelp/help-prepend"));
}

